# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

from Game import GameState
import GridUtils
import LineGame
import random

DEFAULT_NUM_COLORS = 7
DEFAULT_START_PIECES = 5
DEFAULT_NEW_PIECES = 3
DEFAULT_LINE_LENGTH = 5

def add_random_pieces(grid, ncolors, npieces):
    empty = list(GridUtils.enumerate_coords_with_value(grid, 0))
    positions = random.sample(empty, min(npieces, len(empty)))
    for p in positions:
        grid[p] = random.randint(1, ncolors)
    return positions


class LinesGameState(GameState):
    def __init__(self, dimensions=None, params=None, players=2, reset=True):
        GameState.__init__(self, dimensions, params, 1, reset)
        
    def player_score(self, pnum):
        return self.extra_state['score'] if pnum==1 else 0
        
    def num_colors(self):
        return self.game_params.get('num_colors', DEFAULT_NUM_COLORS)
        
    def num_new_pieces(self):
        return self.game_params.get('num_new_pieces', DEFAULT_NEW_PIECES)
        
    def line_length(self):
        return self.game_params.get('line_length', DEFAULT_LINE_LENGTH)
        
    def reset(self):
        GameState.reset(self)
        # keep track of the score and number of consecutive pieces removed, larger combos score higher
        self.extra_state = {'score': 0, 'combo_pieces': 0}
        add_random_pieces(self.grid, self.num_colors(), DEFAULT_START_PIECES)
        
    def remove_lines(self, positions):
        has_lines = False
        for pos in positions:
            for line in LineGame.lines_through_position(self.grid, pos, self.line_length()):
                has_lines = True
                for lp in line:
                    # first N consecutive removed pieces score 1, 2nd N score 2, etc
                    self.grid[lp] = 0
                    self.extra_state['score'] += (self.extra_state['combo_pieces']//self.line_length()) + 1
                    self.extra_state['combo_pieces'] += 1
        return has_lines
        
    def successors(self, ai=False):
        # map empty cells to positions reachable by moving orthogonally through other empty cells        
        empty_groups = {}
        for group in GridUtils.connected_groups(self.grid, 0):
            for pos in group:
                empty_groups[pos] = group
        
        for pos,val in self.enumerate_all():
            if val>0:
                destinations = set()
                # can move to all positions in empty groups from orthogonal neighbors
                for opos, oval in GridUtils.enumerate_ortho_neighbors(self.grid, pos):
                    destinations.update(empty_groups.get(opos,[]))

                for dstpos in destinations:
                    # unless this is an AI move, don't actually compute the resulting state until we have to
                    move = (pos, dstpos)
                    nextstate = self.next_state_for_move(move) if ai else None
                    yield (move, nextstate)
             
    # this overrides GameState.next_state_for_move, hopefully ok       
    def next_state_for_move(self, move):
        pos, dstpos = move
        nextgrid = self.grid.copy()
        nextgrid[pos] = 0
        nextgrid[dstpos] = self.grid[pos]
        nextstate = self.next_state_with_grid(nextgrid)
        # remove lines if any were created
        has_lines = nextstate.remove_lines([dstpos])
        # add new pieces if no lines cleared
        if not has_lines:
            # clear combo counter (should lines created by added pieces affect it?)
            nextstate.extra_state['combo_pieces'] = 0
            new_positions = add_random_pieces(nextgrid, self.num_colors(), self.num_new_pieces())
            # remove any lines created by the new pieces
            nextstate.remove_lines(new_positions)
        return nextstate

    def successors_for_ai(self):
        return self.successors(ai=True)

    def is_game_over(self):
        # game over if no empty cells; the superclass method works but this improves performance
        return 0 not in self.grid

    def animations_for_move(self, move, newstate):
        import PGPieceAnimation
        CELL_MOVE_TIME = 0.1
        # animate move by determining valid path between start and end positions
        path = GridUtils.path_between_positions(self.grid, move[0], move[1], lambda pos,val:val==0)        
        # create segments for move based on path
        move_segments = []
        for index,pos in enumerate(path[:-1]):
            seg = PGPieceAnimation.AnimationSegment(action=PGPieceAnimation.MOVE, 
                                                    from_position=pos,
                                                    to_position=path[index+1],
                                                    start_time=CELL_MOVE_TIME*index,
                                                    end_time=CELL_MOVE_TIME*(index+1))
            move_segments.append(seg)
        
        move_anim = PGPieceAnimation.Animation(position=move[0], segments=move_segments)
        moveend = move_anim.end_time()
        animations = [move_anim]
        # animate new and removed pieces
        for pos in self.itercoords():
            # new if the source cell is refilled, destination isn't new
            if (pos==move[0] or self.grid[pos]==0) and newstate.grid[pos]>0 and pos!=move[1]:
                animations.append(PGPieceAnimation.new_piece_animation(pos, newstate.grid[pos], start_time=moveend))
            # source isn't removed
            if pos!=move[0] and self.grid[pos]>0 and newstate.grid[pos]==0:
                animations.append(PGPieceAnimation.remove_animation(pos, start_time=moveend))
            # if destination is part of a line, add a "remove" segment to the initial move animation
            if pos==move[1] and newstate.grid[pos]==0:
                move_anim.segments.append(PGPieceAnimation.AnimationSegment(action=PGPieceAnimation.REMOVE, start_time=moveend, end_time=moveend+PGPieceAnimation.DEFAULT_DURATION))
                
        return PGPieceAnimation.AnimationGroup(animations)

                    
